use crate::{jwt, schema};
use microrm::prelude::*;
use serde::{Deserialize, Serialize};

type Request = tide::Request<super::ServerStateWrapper>;

#[derive(serde::Serialize)]
pub enum OIDCErrorType {
    InvalidRequest,
    UnauthorizedClient,
    AccessDenied,
    UnsupportedResponseType,
    InvalidScope,
    ServerError,
    TemporarilyUnavailable,
}

/// error type,
pub struct OIDCError<'a>(OIDCErrorType, String, Option<&'a str>);

impl<'a> OIDCError<'a> {
    fn to_response(self) -> tide::Response {
        #[derive(Serialize)]
        struct ErrorOut<'a> {
            error: OIDCErrorType,
            error_description: String,
            state: Option<&'a str>,
        }

        let eo = ErrorOut {
            error: self.0,
            error_description: self.1,
            state: self.2,
        };

        tide::Response::builder(400)
            .body(serde_json::to_vec(&eo).unwrap())
            .build()
    }
}

#[derive(Deserialize)]
struct AuthorizeQueryParams {
    response_type: String,
    client_id: String,
    redirect_uri: String,
    scope: Option<String>,
}

fn do_code_authorize(
    request: Request,
    qp: AuthorizeQueryParams,
    state: Option<&str>,
    client: microrm::WithID<schema::Client>,
) -> Result<tide::Response, OIDCError> {
    todo!()
}

fn do_token_authorize(
    request: Request,
    qp: AuthorizeQueryParams,
    state: Option<&str>,
    client: microrm::WithID<schema::Client>,
) -> Result<tide::Response, OIDCError> {
    let qi = request.state().core.pool.query_interface();

    let shelper = super::session::SessionHelper::new(&request);

    let sauth = shelper
        .get_session(&request)
        .and_then(|sid| shelper.get_auth_for_session(shelper.get_realm().unwrap(), sid));

    if sauth.is_none() {
        // if we don't have any relevant auth info, redirect to login
        let mut login_url = request.url().join("../session/login").unwrap();
        login_url
            .query_pairs_mut()
            .clear()
            .append_pair("redirect", request.url().as_str());
        return Ok(tide::Redirect::new(login_url).into());
    }
    let sauth = sauth.unwrap();
    let user = qi
        .get()
        .by_id(&sauth.user)
        .one()
        .expect("couldn't query db")
        .ok_or_else(|| {
            OIDCError(
                OIDCErrorType::ServerError,
                "user no longer exists".to_string(),
                state,
            )
        })?;

    let issuer = format!(
        "{}/{}",
        request.state().core.config.base_url,
        request.param("realm").unwrap()
    );

    let iat = std::time::SystemTime::now();
    let exp = iat + std::time::Duration::from_secs(request.state().core.config.auth_token_expiry);

    let token = jwt::JWTData {
        sub: user.username.as_str(),
        iss: issuer.as_str(),
        aud: client.shortname.as_str(),
        iat: iat.duration_since(std::time::UNIX_EPOCH).unwrap().as_secs(),
        exp: exp.duration_since(std::time::UNIX_EPOCH).unwrap().as_secs(),
        extras: Default::default(),
    };

    let response_body = serde_json::json!({
        "token": token,
    });

    Ok(tide::Response::builder(200)
        .content_type(tide::http::mime::JSON)
        .body(response_body)
        .build())
}

fn do_authorize(request: Request, state: Option<&str>) -> Result<tide::Response, OIDCError> {
    let qp: AuthorizeQueryParams = request
        .query()
        .map_err(|x| OIDCError(OIDCErrorType::InvalidRequest, x.to_string(), state))?;

    // verify the realm and client_id and redirect_uri
    let qi = request.state().core.pool.query_interface();
    let realm = qi
        .get()
        .by(schema::Realm::Shortname, &request.param("realm").unwrap())
        .one()
        .expect("couldn't query db");
    if realm.is_none() {
        return Err(OIDCError(
            OIDCErrorType::InvalidRequest,
            "No such realm!".to_string(),
            state,
        ));
    }
    let realm = realm.unwrap();

    let client = qi
        .get()
        .by(schema::Client::Realm, &realm.id())
        .by(schema::Client::Shortname, &qp.client_id)
        .one()
        .expect("couldn't query db");
    if client.is_none() {
        return Err(OIDCError(
            OIDCErrorType::UnauthorizedClient,
            "Client does not exist".to_string(),
            state,
        ));
    }

    if qp.response_type == "code" {
        do_code_authorize(request, qp, state, client.unwrap())
    } else if qp.response_type == "token" {
        do_token_authorize(request, qp, state, client.unwrap())
    } else {
        return Err(OIDCError(
            OIDCErrorType::UnsupportedResponseType,
            "Only code and token are understood.".to_string(),
            state,
        ));
    }
}

async fn authorize(request: Request) -> tide::Result<tide::Response> {
    #[derive(Deserialize)]
    struct State {
        state: Option<String>,
    }
    let state: Option<String> = request.query::<State>().ok().map(|x| x.state).flatten();

    match do_authorize(request, state.as_ref().map(|x| x.as_str())) {
        Ok(r) => Ok(r),
        Err(e) => Ok(e.to_response()),
    }
}

#[derive(Deserialize)]
struct TokenRequestBody {
    grant_type: String,
    refresh_token: Option<String>,
    scope: Option<String>,
    redirect_uri: Option<String>,
    code: Option<String>,
    client_id: Option<String>,

    // direct grant
    username: Option<String>,
    password: Option<String>,
}

async fn token(mut request: Request) -> tide::Result<tide::Response> {
    let body: Result<TokenRequestBody, _> = request.body_form().await;
    // let body : TokenRequestBody = request.body_form();
    todo!()
}

pub(super) fn oidc_v1_server(mut route: tide::Route<super::ServerStateWrapper>) {
    route.at("authorize").get(authorize).post(authorize);
    route.at("token").post(token);
}
